package lk;

import lk.datastructures.*;
import java.util.Arrays;

import tsp.TSP;
import tsp.Tour;

/** Implementation of the Lin-Kernighan algorithm based on:
 * 
 * @misc{ applegate1999 ,
 *   author = "D. Applegate and R. Bixby and V. Chv\'atal and W. Cook",
 *   title  = "Finding tours in the {TSP}",
 *   text   = "D. Applegate, R. Bixby, V. Chv'atal, and W. Cook (1999). Finding tours
 *             in the TSP, Report Number 99885, Research Institute for Discrete Mathematics,
 *             Universitaet Bonn.",
 *   year   = "1999",
 *   url    = "www.tsp.gatech.edu/methods/papers/lk_report.ps" 
 * }
 *
 */
public class LinKernighan {

	/** Constant for numerical zero. */
	public static final double NULL = 1E-6;

	/** Mmaximal depth of the serach tree.  */
	public static final int MAXDEPTH = 20;
	
	/** Maximal node degree of the neighborhood graph. */
	public static final int NEIGHBORS = 15;
	
	/** Flag whether to use the bentley marking scheme. */
	public static final boolean BENTLEY = true;
	
	/** Flag whether to use Mak-Martion-Moves. */
	public static final boolean MAKMORTON = true;

	/** Parameter set for the breath of the search */
	public static final int[] breath = { 15, 10, 5 }; // { 15 , 10, 5 };

	static {
		
		System.out.println("\n-------------------------------------------");
		System.out.println("LinKernighan-Settings");
		System.out.println("NEIGHBORS       = " + NEIGHBORS);
		System.out.println("breath          = " + Arrays.toString(breath));
		System.out.println("MAXDEPTH        = " + MAXDEPTH);
		System.out.println("BENTLEY         = " + BENTLEY);
		System.out.println("MAKMORTON       = " + MAKMORTON);
		System.out.println("ITERATIONS      = size /10");
		System.out.println("-------------------------------------------\n");
	}

	/** Queue for the marked nodes without priorities */
	private Queue queue;

	/** Tour data structure that implements next(), prev() and flip(a,b). */
	private LKTour currentTour;
	
	/** The Traveling Salesman problem itself. */
	private TSP tsp;

	/** Dimension of the TSP. */
	private int n;
	
	/** Neighborhood graph. */
	private Neighborhood neighbors;

	/** Constuctor for an instance of the algorithm. */
	public LinKernighan(Tour tour, TSP tsp, Neighborhood neighbors) {

		if (tour == null || tour.size() == 0) {
			System.err.println("could not find inital tour ... ");
			System.exit(1);
		}

		this.currentTour = new LKTour(tsp,tour);

		this.n = currentTour.size();
		
		this.tsp = tsp;

		this.neighbors = neighbors;

		this.deleted = new boolean[n][n];
		this.added = new boolean[n][n];
		this.queue = new Queue(n);
	}

	/** Main algorithm function. */
	public void linKernighan() {
		
		flipSequence = new FlipSequence();

		int v = -1;

		while ((v = queue.pop()) != -1) {

			// call lkSearch( v , currentTour )
			boolean improvedFlipSequenceFound = lkSearch(v);

			
			if (improvedFlipSequenceFound) {
				
				while ( ( flipSequence.nextFlip()) != null ) {	}

				// if we do not use the Bentley-Marking-Scheme, 
				// we use a simple queue
				if ( !BENTLEY ){
					queue.init();
				}
					
				// visualize if we are in gui-mode
				if ( TSP.GUI ) {
					TSP.gui.getJvDisp().addTour( currentTour.getTour() );
				}
			}
		}

	}

	/** The flip sequence itself. */
	private FlipSequence flipSequence;
	
	/** The current base node. */
	private int base;
	
	/** the current best gain at the current recursion node of the search tree. */
	private double bestGain;

	/** Searches for an improving flip sequence starting at node v. */
	private boolean lkSearch(int v) {

		bestGain = 0.0;

		base = v;

		int nextBase = currentTour.next(base);

		markEdgeAsDeleted(base, nextBase);
		
		double delta = 0.0;
		
		// call step( 1 , 0 ) 
		boolean foundImprovement = step(1, delta);

		unMarkEdgeAsDeleted(base, nextBase);

		if (bestGain > NULL) {
			
			if (BENTLEY)
			{
				queue.add(base);
				queue.add(nextBase);
			}
		}

		if (foundImprovement) {
			return true;
		}

		return false;
	}

	/** Recursive implementation of the search of improving flips. */
	private boolean step(int level, double delta) {
		
		// create the lk-ordering for base 
		LkOrdering lkOrdering = new LkOrdering(tsp, base, delta, added,
				deleted, currentTour, neighbors, MAKMORTON && (level == 1));

		// initialize i = 1; 
		int i = 1;

		// while there are unprocessed nodes in the lk-ordering 
		// and  i <= breath(level) ..
		Move move = null;

		while ((move = lkOrdering.nextMove()) != null && i <= breath(level)) {
			
			// let a be the next node in the current lk-ordering
			int nextBase = next(base);

			int a = move.a;

			boolean foundImprovementAtThisLevel = false;
			boolean foundImprovementAtDeeperLevel = false;
			double gain = 0;
			
			// TODO implement execution of the flip operations
			// based on the node nextBase and the node a
			//
			// Implementation of Mak-Morton moves is not mandatory, 
			// and thus optional for the exercise.
			

			if (MAKMORTON && move.isMakMortonMove()) {
				
				// ********************************************************************
				// write code here for the Mak-Morton moves (optional for the exercise)
				// ********************************************************************
				
			} else {
				
				// ********************************************************************
                // write code here for the usual 2-opt moves (mandatory for the exercise)
				//
				// Hints:
				//
				// - use function double c( a, b ); to compute the cost of an edge {a,b}
				//
				// - use move.prev_a to get the predecessor of the a given by the move
				//
				// - use the member variable bestGain to store the best gain found so far
				//
				// - use currentTour.flip(...); to execute the flip operation on the current tour
				// - use flipSequence.addFlips(...); to store the executed flips 
				//  
				// - use markEdgeAsAdded(a, b); to mark the edge {a,b} as added
				// - use markEdgeAsDeleted(a, b); to mark the edge {a,b} as deleted
				//
				// - use 
				//     foundImprovementAtDeeperLevel = step( ... );
				//   to dive along the search tree
				//
				// - use unMarkEdgeAsAdded( a, b ); to remove the marking of the edge {a,b}
				// - use unMarkEdgeAsDeleted( a, b ); to remove the marking of the edge {a,b}
				// ********************************************************************
			}

			if (foundImprovementAtThisLevel || foundImprovementAtDeeperLevel) {

				lkOrdering = null; // help the garbage collector

				if (BENTLEY) {
					queue.add(a);

					queue.add(move.isMakMortonMove() ? move.next_a
							: move.prev_a);
				}

				return true;
				
			} else {
				
				// ********************************************************************
				// write code here for removing the last flip executed 
				//
				// Hint:
				// - use unflip() to remove the last flip

				++i;
				// ********************************************************************
			}
		}

		return false;
	}

	/** Shortcut member function for the next node of v in the current tour. */
	private int next(int v) {
		return currentTour.next(v);
	}

	/** Short member function for removing the last n flips that was added before. */
	private void unFlip( ) {

		Flip lastFlip = flipSequence.lastFlip();
		assert lastFlip != null;
		currentTour.flip(lastFlip.y, lastFlip.x);

	}

	/** Stores which edges are deleted in the current search tree. */
	private boolean[][] deleted;

	/** Stores which edges are added in the current search tree. */
	private boolean[][] added;

	/** Marks the edge {a,b} as deleted. */
	private void markEdgeAsDeleted(int a, int b) {
		deleted[a][b] = deleted[b][a] = true;
	}

	/** Marks the edge {a,b} as added. */
	private void markEdgeAsAdded(int a, int b) {
		added[a][b] = added[b][a] = true;
	}

	/** Unmarks the edge {a,b} as deleted. */
	private void unMarkEdgeAsDeleted(int a, int b) {
		deleted[a][b] = deleted[b][a] = false;
	}

	/** Unmarks the edge {a,b} as added. */
	private void unMarkEdgeAsAdded(int a, int b) {
		added[a][b] = added[b][a] = false;
	}

	/** Returns the maximal breath of the search tree at the given level. */
	private int breath(int level) {
		return breath.length < level ? (level <= MAXDEPTH ? 1 : 0)
				: breath[level - 1];
	}

	/** Starts the algorithm and returns the best tour found so far. */
	public Tour run() {

		linKernighan();
		
		return currentTour.getTour();
	}

	/** Shortcut member function for the objective. */
	public int c(int i, int j) {
		return tsp.costMatrix[i][j];
	}
}