package lk.datastructures;

import java.io.Serializable;
import java.util.Iterator;

import main.Util;

import tsp.TSP;
import tsp.Tour;

/** Tour data structure that implements all necessary operations for 
 * the Lin-Kernighan heuristic. This is the algorithmic representation of 
 * the TSP tour. */
public class LKTour implements Cloneable, Serializable, Iterable<Vertex> {
	
	/** Vector of vertices. */
	private Vertex[] vertices;

	/** TSP instance itself. */
	private TSP tsp;

	/** Constructor for initializing a LKTour from a TSP tour. */
	public LKTour(TSP tsp, Tour tour) {

		this.tsp = tsp;

		int n = tour.size();

		vertices = new Vertex[n];

		for (int i = 0; i < n; ++i) {
			vertices[i] = new Vertex(tour.elementAt(i));
		}

		for (int i = 0; i < n; ++i) {
			Vertex prev = vertices[i];
			Vertex next = vertices[(i + 1) % n];

			prev.setNext(next);
			next.setPrev(prev);
		}
	}

	/** Returns the node with the given id. */
	public Vertex get(int id) {
		return vertices[id];
	}

	/** Returns the successor of the node with the given id. */
	public int next(int id) {
		return get(id).next().id();
	}

	/** Returns the predecessor of the node with the given id. */
	public int prev(int id) {
		return get(id).prev().id();
	}

	/** Executes a flip operation on that tour. */
	public double flip(int ida, int idb) {
		Vertex a = get(ida);
		Vertex b = get(idb);

		assert a.prev() != b; // senseless: like prev(a) != nextBase 
		assert a.prev() != b.next(); // senseless: like a != base 

		if (a == b)
		{
			return 0.0;
		}

		Vertex prevA = a.prev(), nextB = b.next();

		reverse(prevA, a, b, nextB);

		return c(prevA, a) + c(b, nextB) - c(prevA, b) - c(a, nextB);
	}

	/** Inverts a path defined by a and b in that tour. */
	private void reverse(Vertex prevA, Vertex a, Vertex b, Vertex nextB) { // System.out.println("reverse("+prevA.id()+","+a.id()+","+b.id()+","+nextB.id()+")");

		// input:  ... prevA < a ... b > nextB ...
		// output: ... prevA < b ... a > nextB ...

		if (a == b)
		{
			return;
		}

		Vertex v = prevA;
		Vertex w = b;

		// starting with b we append all nodes on prevA (until a is reached) 

		while (true) // APPEND
		{
			Vertex vorW = w.prev();

			connect(v, w);

			if (w == a)
				break;

			v = w;
			w = vorW;
		}

		// connect the segments
		connect(a, nextB);
	}

	/** Connects two vertices. */
	private void connect(Vertex a, Vertex b) {
		a.setNext(b);
		b.setPrev(a);
	}

	/** Inner iterator class for iterating this tour. */
	private class I implements Iterator<Vertex> {
		Vertex next = null;
		int size;
		int rank;
		boolean reverse;

		public I(Vertex start, int size, boolean reverse) {
			this.next = start;
			this.size = size;
			this.rank = 0;
			this.reverse = reverse;
		}

		public boolean hasNext() {
			return (rank < size);
		}

		public Vertex next() {
			Vertex ret = next;
			next = reverse ? ret.prev() : ret.next();
			++rank;
			return ret;
		}

		public void remove() {
		}
	}

	public Iterator<Vertex> iterator() {
		return new I(vertices[0], size(), false);
	}

	/** Returns the Vertex at position i of this tour. */
	public Vertex at(int i) {
		int j = 0;
		for (Vertex v : this)
			if (j++ == i)
				return v;

		return null;
	}

	/** Returns the size of this tour. */
	public int size() {
		return vertices.length;
	}

	private static final long serialVersionUID = 1L;

	/** Returns the data representation of this tour. */
	public tsp.Tour getTour() {

		Tour tour = new Tour();

		for (Vertex v : this) {
			tour.add(v.getNode());
		}

		return tour;
	}

	/** Computes the cost of this tour. */
	public double cost() {

		double cost = 0.0;

		int n = vertices.length;

		for (int i = 0; i <= n; ++i) {
			cost += tsp.costMatrix[vertices[i % n].id()][vertices[(i + 1) % n]
					.id()];
		}
		return cost;
	}

	/** Shortcut member function for the objective. */
	private int c(Vertex a, Vertex b) {
		return tsp.costMatrix[a.id()][b.id()];
	}

	/** Returns a deep copy of this tour. */
	public LKTour clone() {
		return (LKTour) Util.deepCopy(this);
	}
}