package tsp;

import gui.TSPGui;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Vector;

import main.Util;

/** Represents a TSP instance. */
public class TSP {

	
	/** Flag whether a GUI should be started. */
	public static final boolean GUI = true ;
	
	/** Graphical user interface for the traveling salesman problem. */
	public static TSPGui gui = null;
	
	/** Name of the file that was read. */
	public static String  FNAME = null;
	
	/** Name of the instance. */
	public static String  NAME = "";
	
	/** Comment for the instance. */
	public static String  COMMENT = "";
	
	/** Cype of the problem */
	public static String  TYPE = "";

	/** Dimenstion of the instance. */
	public static int     DIMENSION = -1;
	
	/** Type of the edege weights. */
	public static String  EDGE_WEIGHT_TYPE = "";

	/** Vector of nodes forming the instance. */
	private Vector<Node> nodes = new Vector<Node>();

	/** Maximal x-coordinate. */
	private double maxX = 0;
	
	/** Maximal y-coordinate. */
	private double maxY = 0;
	
	/** Minimal x-coordinate. */
	private double minX = Double.MAX_VALUE;
	
	/** Minimal y-coordinate. */
	private double minY = Double.MAX_VALUE;
	
	/** Complete cost matrix of the instance. */
	public int[][] costMatrix = null;

	/** Returns the id of the given node. */
	public int getId(Node node) 
	{
		int i = 0;
		for ( Node m : nodes ) {
			if (m == node)
				return i++;
		}
		return -1;
	}

	/** Returns the node with the given id. */
	public Node get(int id) {
		return nodes.elementAt(id - 1);
	}
	

	/** Constructor. */
	public TSP( String fname ) {
		
		FNAME = fname;
		
		init( );
		
		costMatrix = new int[DIMENSION][DIMENSION];
		
		// initialize the cost matrix
		for( Node tail : nodes )
		{
			for( Node head : nodes )
			{
				costMatrix[tail.getID()][head.getID()] = 
						Util.cost( tail, head );
			}
			
		}
		
		assert DIMENSION == nodes.size() ;
		
		printInfos();
	}

	/** Reads the input file. */
	private void init() {
		
		// flags whether to parse the node coordinates section of the instance
		// file
		boolean readCoordinates = false;
		
		int id = 0;

		// read the input file
		try {

			FileInputStream is = new FileInputStream(FNAME);

			InputStreamReader isr = new InputStreamReader(is, "UTF8");
			BufferedReader f = new BufferedReader(isr);

			String line;

			while ((line = f.readLine()) != null) {

				// to enable printing of each line to read
				// System.out.println(" read:>\t" + line);

				if (line.startsWith("EOF") || line.matches("^ *$") ) {
					break;
				}
				
				// ignore empty lines
				if ( line.matches("^ *$") ) {
					continue;
				}

				// ignore comments
				if (line.startsWith("#")) {
					continue;
				}

				if (line.startsWith("NAME")) {
					NAME = line.split(":")[1].trim();
					continue;
				}

				if (line.startsWith("COMMENT")) {
					COMMENT = line.split(":")[1].trim();
					continue;
				}

				if (line.startsWith("TYPE")) {
					TYPE = line.split(":")[1].trim();
					
					if( ! TYPE.equals("TSP") )
					{
						System.err.println("can not handle problem type >" + TYPE+ "<");
						System.exit( 1 );
					}
					continue;
				}

				if (line.startsWith("DIMENSION")) {
					DIMENSION = Integer.parseInt(line.split(":")[1].trim());
					continue;
				}

				if (line.startsWith("EDGE_WEIGHT_TYPE")) {
					EDGE_WEIGHT_TYPE = line.split(":")[1].trim();
					
					if( ! EDGE_WEIGHT_TYPE.equals( "EUC_2D")  )
					{
						System.err.println("can not handle edge weight type >" + EDGE_WEIGHT_TYPE + "<");
						System.exit( 1 );
					}

					continue;
				}

				if (line.startsWith("NODE_COORD_SECTION")) {
					readCoordinates = true;
					continue;
				}

				if (readCoordinates) {
					
					try {
						String[] items = line.trim().split(" +");

						if (items.length != 3) {
							System.err.println("can not read line >" + line + "<\n");
							System.exit(1);
						}

						String name = "" + Integer.parseInt(items[0].replace(',', '.'));
						
						double x = Double.parseDouble(items[1].replace(',', '.'));
						double y = Double.parseDouble(items[2].replace(',', '.'));

						nodes.add(new Node(id++, name, x, y));

						maxX = x > maxX ? x : maxX;
						maxY = y > maxY ? y : maxY;
						
						minX = x < minX ? x : minX;
						minY = y < minY ? y : minY;


					} catch (NumberFormatException e) {
						System.err.println("exception during converting numbers of line: " + line);
						System.exit( 1 );
					}
				}

			}

			f.close();

		} catch (IOException e) {
			System.err.println("error during reading of file " + FNAME + "\n" + e);
			Util.printUsageAndExit();
		}
		
		// traverse all coordinates to positive octant to have all values non-negative (especially needed for delaunay triangulation)
		
		double xOffset = Math.abs( minX );
		double yOffset = Math.abs( minY );
		
		maxX += xOffset;
		maxY += yOffset;
		
		for( Node node : nodes )
		{
			node.setX(node.getX() + xOffset);
			node.setY(node.getY() + yOffset);
		}
	}

	/** Prints some informations about the TSP instance to stdout. */
	private void printInfos() {
		System.out.println("---------------------------------------");
		System.out.println("TSP-Settings");
		System.out.println("FILE             = " + FNAME );
		System.out.println("NAME             = " + NAME);
		System.out.println("COMMENT          = " + COMMENT);
		System.out.println("DIMENSION        = " + DIMENSION);
		System.out.println("EDGE_WEIGHT_TYPE = " + EDGE_WEIGHT_TYPE );
		System.out.println("maxX             = " + maxX );
		System.out.println("maxY             = " + maxY );
		System.out.println("---------------------------------------");
	}

	/** Returns a vector of nodes. */
	public Vector<Node> getNodes() {
		return nodes;
	}

	/** Returns the maximal x-coordinate. */
	public double getMaxX() {
		return maxX;
	}

	/** Returns the maximal y-coordinate. */
	public double getMaxY() {
		return maxY;
	}
	
	/** Returns the name of the input file. */
	public String getFileName() { 
		return FNAME; 
	}
}